import pathlib

import click

from aced_submission import NaturalOrderGroup
from aced_submission.fhir_store import fhir_store
from aced_submission.meta import meta
from aced_submission.uploader import files
import logging


FORMAT = '%(levelname)s:%(message)s'
logging.basicConfig(format=FORMAT, level=logging.INFO)


@click.group(cls=NaturalOrderGroup)
def cli():
    pass


cli.add_command(meta)
cli.add_command(files)
cli.add_command(fhir_store)


@cli.command(name='schema-publish')
@click.argument('dictionary_path', default='iceberg/schemas/gen3/aced.json')
@click.option('--bucket', default="s3://aced-public", help="Bucket target", show_default=True)
@click.option('--production', default=False, is_flag=True, show_default=True,
              help="Write to aced.json, otherwise aced-test.json")
def schema_publish(dictionary_path, bucket, production):
    """Copy dictionary to s3 (note:aws cli dependency)"""

    dictionary_path = pathlib.Path(dictionary_path)
    assert dictionary_path.is_file(), f"{dictionary_path} should be a path"
    click.echo(f"Writing schema into {bucket}")
    dictionary_path.is_file()
    import subprocess
    if production:
        cmd = f"aws s3 cp {dictionary_path} {bucket}/aced.json".split(' ')
    else:
        cmd = f"aws s3 cp {dictionary_path} {bucket}/aced-test.json".split(' ')
    s3_cp = subprocess.run(cmd)
    assert s3_cp.returncode == 0, s3_cp
    print("OK")


if __name__ == '__main__':
    cli()



import pathlib
import pytest

from gen3_tracker.meta.dataframer import LocalFHIRDatabase

from aced_submission.meta_flat_load import load_flat, DEFAULT_ELASTIC


@pytest.fixture()
def project_id() -> str:
    return "test-observations"

@pytest.fixture()
def meta_dir() -> str:
    return "tests/fixtures/Breast_Cancer/META"


def test_load_observations(meta_dir, project_id, tmpdir) -> None:
    """Ensure we can load observations, the ES mapping should reflect updated fields."""
    # change to the temporary directory
    print(pathlib.Path.cwd())
    print(project_id)

    # create a database from META
    work_path = tmpdir
    assert pathlib.Path(work_path).exists(), f"Directory {work_path} does not exist."
    work_path = pathlib.Path(work_path)
    db_path = (work_path / "local_fhir.db")
    db_path.unlink(missing_ok=True)

    db = LocalFHIRDatabase(db_name=db_path)
    db.load_ndjson_from_dir(path=meta_dir)

    load_flat(project_id=project_id,
              generator=db.flattened_observations(),
              index='observation',
              limit=None,
              elastic_url=DEFAULT_ELASTIC,
              output_path=None
              )

    from elasticsearch import Elasticsearch
    elastic = Elasticsearch([DEFAULT_ELASTIC], request_timeout=120)
    # retrieve by alias
    alias_mapping = elastic.indices.get_mapping(index='observation')
    # retrieve by index name
    index_mapping = elastic.indices.get_mapping(index='gen3.aced.io_observation_0')
    assert alias_mapping == index_mapping, "Alias mapping should match index mapping."
    # retrieve array config
    array_config_mapping = elastic.indices.get_mapping(index='gen3.aced.io_observation-array-config_0')
    assert array_config_mapping, "Array config mapping should exist."
    # no fields are arrays currently
    results = elastic.search(index='gen3.aced.io_observation-array-config_0')
    print(results)
    assert len(results['hits']['hits']) > 0, "Expected to find array config."
    assert len(results['hits']['hits'][0]['_source']['array']) == 2, "Expected to find 2 array fields."

    # aggregate by auth_resource_path
    # Define the query
    program, project = project_id.split('-')
    query = {
        "query": {
            "match": {
                "auth_resource_path": f"/programs/{program}/projects/{project}"
            }
        }
    }

    # Count the documents
    count = elastic.count(index="observation", body=query)

    assert count['count'] > 0, "Expected to find some observations."

    #
    # now lets make sure when we add a new field, it gets added to the mappingq
    #

    # add a new observation, with a new field `patient_us_core_favorite_ice_cream`
    observation = {"resourceType": "Observation", "id": "5f5454dd-XXXX-4bd9-YYYY-b4318ae469c3", "status": "final",
                   "category": "Laboratory",
                   "code": "Progesterone receptor Ag [Presence] in Breast cancer specimen by Immune stain",
                   "subject": "Patient/52e0c68f-8f6d-42f4-922c-0e32559d5a2b",
                   "effectiveDateTime": "2004-11-12T17:33:16-05:00", "issued": "2004-11-12T17:33:16.515-05:00",
                   "identifier": None, "value_normalized": "Positive (qualifier value)", "value_numeric": None,
                   "patient": "52e0c68f-8f6d-42f4-922c-0e32559d5a2b",
                   "patient_us_core_ethnicity": "Non Hispanic or Latino",
                   "patient_us_core_favorite_ice_cream": "Vanilla",
                   "patient_us_core_race": "Black or African American", "patient_us_core_birthsex": "F",
                   "project_id": "test-observations", "auth_resource_path": "/programs/test/projects/observations"}

    load_flat(project_id=project_id,
              generator=[observation],
              index='observation',
              limit=None,
              elastic_url=DEFAULT_ELASTIC,
              output_path=None
              )

    # retrieve by alias
    alias_mapping = elastic.indices.get_mapping(index='observation')
    assert alias_mapping['gen3.aced.io_observation_0']['mappings']['properties']['patient_us_core_favorite_ice_cream'] == {'type': 'keyword'}, "Expected to find the new field in the mapping."

    # we should find it by id
    doc = elastic.get(index='observation', id='5f5454dd-XXXX-4bd9-YYYY-b4318ae469c3')
    assert doc['_source']['patient_us_core_favorite_ice_cream'] == "Vanilla", "Expected to find the new field in the document."

    #
    # test the array_fields
    #
    # add a new observation, with a new field that is an array `patient_us_core_favorite_colors`
    observation = {"resourceType": "Observation", "id": "5f5454dd-ZZZZ-4bd9-DDDDD-b4318ae469c3", "status": "final",
                   "category": "Laboratory",
                   "code": "Progesterone receptor Ag [Presence] in Breast cancer specimen by Immune stain",
                   "subject": "Patient/52e0c68f-8f6d-42f4-922c-0e32559d5a2b",
                   "effectiveDateTime": "2004-11-12T17:33:16-05:00", "issued": "2004-11-12T17:33:16.515-05:00",
                   "identifier": None, "value_normalized": "Positive (qualifier value)", "value_numeric": None,
                   "patient": "52e0c68f-8f6d-42f4-922c-0e32559d5a2b",
                   "patient_us_core_ethnicity": "Non Hispanic or Latino",
                   "patient_us_core_favorite_ice_cream": "Vanilla",
                   "patient_us_core_favorite_colors": ["Red", "Green"],
                   "patient_us_core_race": "Black or African American", "patient_us_core_birthsex": "F",
                   "project_id": "test-observations", "auth_resource_path": "/programs/test/projects/observations"}

    load_flat(project_id=project_id,
              generator=[observation],
              index='observation',
              limit=None,
              elastic_url=DEFAULT_ELASTIC,
              output_path=None
              )

    # retrieve by alias
    alias_mapping = elastic.indices.get_mapping(index='observation')
    assert alias_mapping['gen3.aced.io_observation_0']['mappings']['properties']['patient_us_core_favorite_colors'] == {'type': 'keyword'}, "Expected to find the new field in the mapping."

    # we should find it by id
    doc = elastic.get(index='observation', id='5f5454dd-ZZZZ-4bd9-DDDDD-b4318ae469c3')
    assert doc['_source']['patient_us_core_favorite_colors'] == ["Red", "Green"], "Expected to find the new field in the document."

    results = elastic.search(index='gen3.aced.io_observation-array-config_0')
    print(results)
    assert len(results['hits']['hits']) > 0, "Expected to find array config."
    assert 'patient_us_core_favorite_colors' in results['hits']['hits'][0]['_source']['array']




#!/usr/bin/env python3
import datetime
import json
import logging
import pathlib
import sys
import urllib
import uuid
from dataclasses import dataclass
from itertools import islice
from multiprocessing.pool import Pool
from time import sleep
from typing import Iterator

import click
import jwt
import requests
from gen3.auth import Gen3Auth
from gen3.file import Gen3File
from gen3.index import Gen3Index
from orjson import orjson
from tqdm import tqdm

from aced_submission import NaturalOrderGroup

logger = logging.getLogger(__name__)

LOGGED_ALREADY = set({})

ACED_CODEABLE_CONCEPT = uuid.uuid3(uuid.NAMESPACE_DNS, 'aced.ipd/CodeableConcept')
ACED_NAMESPACE = uuid.uuid3(uuid.NAMESPACE_DNS, 'aced-ipd.org')


@dataclass
class UploadResult:
    """Results of Upload DocumentReference."""
    document_reference: dict
    """The source document reference."""
    elapsed: datetime.timedelta
    """Amount of time it took to upload."""
    exception: Exception = None
    """On error."""


def _chunk(arr_range, arr_size):
    """Iterate in chunks."""
    arr_range = iter(arr_range)
    return iter(lambda: tuple(islice(arr_range, arr_size)), ())


@click.group(cls=NaturalOrderGroup)
def files():
    """Project files (Omics, Imaging, ...)."""
    pass


def extract_endpoint(gen3_credentials_file):
    """Get base url of jwt issuer claim."""
    with open(gen3_credentials_file) as fp:
        api_key = json.load(fp)['api_key']
        claims = jwt.decode(api_key, options={"verify_signature": False})
        assert 'iss' in claims
        return claims['iss'].replace('/user', '')


def _upload_document_reference(document_reference: dict, bucket_name: str,
                               program: str, project: str, duplicate_check: bool, credentials_file: str,
                               source_path: str) -> UploadResult:
    """Write a single document reference to indexd and upload file."""

    try:
        start = datetime.datetime.now()
        # print(('starting', document_reference['id'], start.isoformat()))

        file_client, index_client = _gen3_services(credentials_file)

        attachment, md5sum, source_path_extension = _extract_extensions(document_reference)

        source_path = _extract_source_path(attachment, source_path, source_path_extension)

        file_name = source_path.lstrip('./').lstrip('file:///')
        object_name = attachment['url'].lstrip('./').lstrip('file:///')

        metadata = _update_indexd(attachment, bucket_name, document_reference, duplicate_check, index_client, md5sum,
                                  object_name, program, project)

        # create a record in gen3 using document_reference's id as guid, get a signed url
        # SYNC

        document = file_client.upload_file_to_guid(guid=document_reference['id'], file_name=object_name, bucket=bucket_name)
        assert 'url' in document, document
        signed_url = urllib.parse.unquote(document['url'])
        file_name = pathlib.Path(file_name)
        assert file_name.exists(), f"{file_name} does not exist"

        _upload_file_to_signed_url(file_name, md5sum, metadata, signed_url)

        end = datetime.datetime.now()
        # print(('complete', document_reference['id'], end.isoformat(), end-start, attachment["size"]))

        return UploadResult(document_reference, end - start)
    except Exception as e:  # noqa
        return UploadResult(document_reference, None, e)


def _read_in_chunks(file_object, chunk_size):
    """Iterator to read file in chunks"""
    while True:
        data = file_object.read(chunk_size)
        if not data:
            break
        yield data


def _upload_file_to_signed_url(file_name, md5sum, metadata, signed_url):
    """Upload file """

    # When you use this header, Amazon S3 checks the object against the provided MD5 value and,
    # if they do not match, returns an error.

    # content_md5 = base64.b64encode(bytes.fromhex(md5sum))
    # headers = {'Content-MD5': content_md5}
    # # attach our metadata to s3 object
    # for key, value in metadata.items():
    #     headers[f"x-amz-meta-{key}"] = value

    with open(file_name, 'rb') as fp:
        # SYNC
        response = requests.put(signed_url, data=fp)
        response_text = response.text
        assert response.status_code == 200, (signed_url, response_text)


def _update_indexd(attachment, bucket_name, document_reference, duplicate_check, index_client, md5sum, object_name,
                   program, project):
    hashes = {'md5': md5sum}
    assert 'id' in document_reference, document_reference
    guid = document_reference['id']
    metadata = {
        **{
            'datanode_type': 'DocumentReference',
            'datanode_object_id': guid
        },
        **hashes}
    # SYNC
    existing_record = None
    s3_url = f"s3://{bucket_name}/{guid}/{object_name}"
    if duplicate_check:
        try:
            existing_record = index_client.get_record(guid=document_reference["id"])
        except Exception: # noqa
            pass
        if existing_record:
            skip_delete = all([
                existing_record['hashes']['md5'] == md5sum,
                s3_url in existing_record['urls']
            ])
            if not skip_delete:
                # SYNC
                logger.debug(f"Deleting existing record {document_reference['id']}")
                index_client.delete_record(guid=document_reference["id"])
                existing_record = None
    if not existing_record:
        try:
            _ = index_client.create_record(
                did=document_reference["id"],
                hashes=hashes,
                size=attachment["size"],
                authz=[f'/programs/{program}/projects/{project}'],
                file_name=object_name,
                metadata=metadata,
                urls=[s3_url]  # TODO make a DRS URL
            )
        except (requests.exceptions.HTTPError, AssertionError) as e:
            if not ('already exists' in str(e)):
                raise e
            logger.info(f"indexd record already exists, continuing upload. {document_reference['id']}")
    return metadata


def _extract_source_path(attachment, source_path, source_path_extension) -> str:
    if source_path:
        source_path = pathlib.Path(source_path)
        assert source_path.is_dir(), f"Path is not a directory {source_path}"
        source_path = source_path / attachment['url'].lstrip('./').lstrip('file:///')
        source_path = str(source_path)
    else:
        if len(source_path_extension) == 1:  # "Missing source_path extension."
            source_path = source_path_extension[0]['valueUrl']
        else:
            source_path = attachment['url'].lstrip('./').lstrip('file:///')
    return source_path


def _extract_extensions(document_reference):
    """Extract useful data from document_reference."""
    attachment = document_reference['content'][0]['attachment']
    md5_extension = [_ for _ in attachment['extension'] if
                     _['url'] == "http://aced-idp.org/fhir/StructureDefinition/md5"]
    assert len(md5_extension) == 1, "Missing MD5 extension."
    md5sum = md5_extension[0]['valueString']
    source_path_extension = [_ for _ in attachment['extension'] if
                             _['url'] == "http://aced-idp.org/fhir/StructureDefinition/source_path"]
    return attachment, md5sum, source_path_extension


def _gen3_services(credentials_file: str) -> (Gen3File, Gen3Index):
    """Create Gen3 Services."""
    credentials_file = str(pathlib.Path(credentials_file).expanduser())
    endpoint = extract_endpoint(credentials_file)
    # logger.debug(endpoint)
    auth = Gen3Auth(endpoint, refresh_file=credentials_file)
    file_client = Gen3File(endpoint, auth)
    index_client = Gen3Index(endpoint, auth)
    return file_client, index_client


def document_reference_reader(document_reference_path) -> Iterator[dict]:
    """Read DocumentReference.ndjson file or bundle."""
    if 'ndjson' in document_reference_path:
        with open(document_reference_path) as fp:
            for _ in fp.readlines():
                yield orjson.loads(_)
    else:
        document_reference_path = pathlib.Path(document_reference_path)
        for input_file in document_reference_path.glob('*.json'):
            with open(input_file, "rb") as fp:
                bundle_ = orjson.loads(fp.read())
                if 'entry' not in bundle_:
                    print(f"No 'entry' in bundle {input_file} ")
                    break
            for entry in bundle_['entry']:
                resource = entry['resource']
                if resource['resourceType'] == 'DocumentReference':
                    yield resource


@files.command(name='upload')
@click.option('--bucket_name', show_default=True,
              help='Destination bucket name')
@click.option('--document_reference_path', required=True, default=None, show_default=True,
              help='Path to DocumentReference.ndjson')
@click.option('--source_path', required=False, default=None, show_default=True,
              help='Path on local file system')
@click.option('--program', required=True, show_default=True,
              help='Gen3 program')
@click.option('--project', required=True, show_default=True,
              help='Gen3 project')
@click.option('--credentials_file', default='~/.gen3/credentials.json', show_default=True,
              help='API credentials file downloaded from gen3 profile.')
@click.option('--duplicate_check', default=False, is_flag=True, show_default=True,
              help="Check for existing indexd records")
@click.option('--worker_count', default=10, show_default=True,
              help="Number of worker processes")
@click.option('--silent', default=False, is_flag=True, show_default=True,
              help="No progress bar, or other output")
@click.option('--state_dir', default='~/.gen3/aced-uploader', show_default=True,
              help='Directory for upload status')
@click.option('--ignore_state', default=False, is_flag=True, show_default=True,
              help="Upload file, even if already uploaded")
def upload_document_reference(bucket_name, document_reference_path, source_path, program, project, credentials_file,
                              duplicate_check, worker_count, silent, state_dir, ignore_state):
    """Upload data file associated with DocumentReference.

    """
    _ = pathlib.Path(document_reference_path)
    assert _.is_file() or _.is_dir(), f"{document_reference_path} directory does not exist"

    state_dir = pathlib.Path(state_dir).expanduser()
    state_dir.mkdir(parents=True, exist_ok=True)
    state_file = state_dir / "state.ndjson"

    # all attempts ids are incomplete until they succeed
    incomplete = set()
    # key:document_reference.id of completed file transfers
    completed = set()
    # key:document_reference.id  of failed file transfers
    exceptions = {}

    # for unhandled exceptions: map multiprocessing tasks results to document_reference
    # keyed by result
    document_reference_lookup = {}

    # progress bar control
    document_references_size = 0
    # loop control
    document_references_length = 0

    already_uploaded = set()
    if not ignore_state and state_file.exists():
        with open(state_file, "rb") as fp:
            for _ in fp.readlines():
                state = orjson.loads(_)
                already_uploaded.update([_ for _ in state['completed']])

    for _ in document_reference_reader(document_reference_path):
        if _['id'] not in already_uploaded:
            incomplete.add(_['id'])
            document_references_size += _['content'][0]['attachment']['size']
            document_references_length += 1
        else:
            if not silent:
                print(f"{_['id']} already uploaded, skipping", file=sys.stderr)

    # re-open file, process it a chunk at a time
    with Pool(processes=worker_count) as pool:
        results = []
        for document_reference in document_reference_reader(document_reference_path):
            if document_reference['id'] in already_uploaded:
                continue
            result = pool.apply_async(
                func=_upload_document_reference,
                args=(
                    document_reference,
                    bucket_name,
                    program,
                    project,
                    duplicate_check,
                    credentials_file,
                    source_path
                )
            )
            results.append(result)
            document_reference_lookup[id(result)] = document_reference['id']

        # close the process pool
        pool.close()

        # poll the results every sec.
        with tqdm(total=document_references_size, unit='B', disable=silent,
                  unit_scale=True, unit_divisor=1024) as pbar:
            while True:
                results_to_remove = []
                for record in results:
                    if record.ready() and record.successful():
                        r = record.get()
                        # print(f'ready and successful {id(r)}')
                        if r.exception:
                            exceptions[r.document_reference['id']] = {
                                    'exception': str(r.exception),
                                    'document_reference': {
                                        'id': r.document_reference
                                    }
                                }
                        elif r.document_reference['id'] not in completed:
                            completed.add(r.document_reference['id'])
                            incomplete.remove(r.document_reference['id'])

                        results_to_remove.append(record)
                        document_references_length = document_references_length - 1
                        pbar.set_postfix(file=f"{r.document_reference['id'][-6:]}", elapsed=f"{r.elapsed}")
                        pbar.update(r.document_reference['content'][0]['attachment']['size'])
                        sleep(.1)  # give screen a chance to refresh

                    if record.ready() and not record.successful():
                        print('record.ready() and not record.successful()')
                        # capture exception, we shouldn't get here as all exception should be caught
                        document_reference_id = document_reference_lookup[id(record)]
                        try:
                            record.get()
                        except Exception as e:  # noqa
                            if document_reference_id not in exceptions:
                                exceptions[document_reference_id] = {
                                    'exception': str(e),
                                    'document_reference': {
                                        'id': document_reference_id
                                    }
                                }
                                document_references_length = document_references_length - 1
                                # print('not successful', document_references_length, e)

                if document_references_length == 0:
                    break

                # print(f'sleeping(1) document_references_length: {document_references_length}')
                sleep(1)

                # using list comprehension to cull processed results
                results = [_ for _ in results if _ not in results_to_remove]

        with open(state_file, "a+b") as fp:
            fp.write(orjson.dumps(
                    {
                        'timestamp': datetime.datetime.now().isoformat(),
                        'completed': [_ for _ in completed],
                        'incomplete': [_ for _ in incomplete],
                        'exceptions': exceptions
                    },
                    option=orjson.OPT_APPEND_NEWLINE
                ))

        if not silent:
            print(f"Wrote state to {state_file}", file=sys.stderr)
        if len(incomplete) == 0:
            if not silent:
                print('OK', file=sys.stderr)
            exit(0)
        else:
            if not silent:
                print('Incomplete transfers:', [_ for _ in incomplete], file=sys.stderr)
                print('Errors:', [f"{_}: {exceptions[_]['exception']}" for _ in exceptions], file=sys.stderr)
            exit(1)


if __name__ == '__main__':
    files()



import io
import logging
import pathlib
import uuid
from collections import defaultdict
from datetime import datetime
from itertools import islice
from typing import List

import yaml
from dictionaryutils import DataDictionary, dictionary
from yaml import SafeLoader

import psycopg2
import json
import inflection

from aced_submission.pelican import DataDictionaryTraversal

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
logging.getLogger('elasticsearch').setLevel(logging.WARNING)

LOGGED_ALREADY = []


def _connect_to_postgres():
    """Connect to postgres based on environmental variables."""
    return psycopg2.connect('')


def _init_dictionary(root_dir_=None, dictionary_url=None):
    """Initialize gen3 data dictionary from either directory or url"""
    d = DataDictionary(root_dir=root_dir_, url=dictionary_url)
    dictionary.init(d)
    # the gdcdatamodel expects dictionary initiated on load, so this can't be
    # imported on module level
    from gdcdatamodel import models as md

    return d, md


def _table_mappings(dictionary_path, dictionary_url):
    """Gen3 vertex/edge table mappings."""
    _dictionary, model = _init_dictionary(root_dir_=dictionary_path, dictionary_url=dictionary_url)
    ddt = DataDictionaryTraversal(model)
    desired_keys = [
        '__dst_class__',
        '__dst_src_assoc__',
        '__dst_table__',
        '__label__',
        '__src_class__',
        '__src_dst_assoc__',
        '__src_table__',
        '__tablename__'
    ]

    def _transform(ddt_) -> List[dict]:
        for d in ddt_.get_edges():
            yield {k.replace('_', ''): v for k, v in d.__dict__.items() if k in desired_keys}

    mapping = _transform(ddt)
    return mapping


def chunk(arr_range, arr_size):
    """Iterate in chunks."""
    arr_range = iter(arr_range)
    return iter(lambda: tuple(islice(arr_range, arr_size)), ())


def load_vertices(files, connection, dependency_order, project_id, mapping):
    """Load files into database vertices."""
    logger.info(f"Number of files available for load: {len(files)}")
    for entity_name in dependency_order:
        path = next(iter([fn for fn in files if str(fn).endswith(f"{entity_name}.ndjson")]), None)
        if not path:
            logger.warning(f"No file found for {entity_name} skipping")
            continue
        data_table_name = next(
            iter(
                set([m['dsttable'] for m in mapping if m['dstclass'].lower() == entity_name.lower()] +
                    [m['srctable'] for m in mapping if m['srcclass'].lower() == entity_name.lower()])
            ),
            None)
        if not data_table_name:
            logger.warning(f"No mapping found for {entity_name} skipping")
            continue
        logger.info(f"loading {path} into {data_table_name}")

        columns = ['node_id', '_props', 'acl', '_sysan', 'created']
        # Select only columns to be updated (in my case, all non-id columns)
        # update_set = ", ".join([f"{v}=EXCLUDED.{v}" for v in ['_props', 'acl', '_sysan', 'created']])
        update_set = ' _props=EXCLUDED._props, acl=EXCLUDED.acl, _sysan=EXCLUDED._sysan, created=EXCLUDED.created '

        with connection.cursor() as cursor:
            with open(path) as f:
                # copy a block of records into a file like stringIO buffer
                record_count = 0
                for lines in chunk(f.readlines(), 1000):
                    # Creates temporary empty table with same columns and types as
                    # the final table
                    cursor.execute(
                        f"""
                        CREATE TEMPORARY TABLE tmp_{data_table_name} (LIKE {data_table_name})
                        ON COMMIT DROP
                        """
                    )
                    buf = io.StringIO()
                    for line in lines:
                        record_count += 1
                        d_ = json.loads(line)
                        d_['object']['project_id'] = project_id
                        obj_str = json.dumps(d_['object'])
                        _csv = f"{d_['id']}\t{obj_str}\t{{}}\t{{}}\t{datetime.now()}".replace('\n', '\\n').replace("\\",
                                                                                                                   "\\\\")
                        _csv = _csv + '\n'
                        buf.write(_csv)
                    buf.seek(0)
                    # efficient way to write to postgres
                    cursor.copy_from(buf, f'tmp_{data_table_name}', sep='\t',
                                     columns=columns)
                    # handle conflicts
                    cursor.execute(
                        f"""
                        INSERT INTO {data_table_name}({', '.join(columns)})
                        SELECT  node_id, _props::jsonb, acl, _sysan, created FROM tmp_{data_table_name}
                        ON CONFLICT (node_id) DO UPDATE SET {update_set}
                        """
                    )
                    logger.info(f"wrote {record_count} records to {data_table_name} from {path}")
                    connection.commit()
        connection.commit()


def load_edges(files, connection, dependency_order, mapping, project_node_id):
    """Load files into database edges."""
    logger.info(f"Number of files available for load: {len(files)}")
    for entity_name in dependency_order:
        path = next(iter([fn for fn in files if str(fn).endswith(f"{entity_name}.ndjson")]), None)
        if not path:
            logger.warning(f"No file found for {entity_name} skipping")
            continue

        with connection.cursor() as cursor:
            with open(path) as f:
                # copy a block of records into a file like stringIO buffer
                record_count = 0

                for lines in chunk(f.readlines(), 100):
                    buffers = defaultdict(io.StringIO)
                    for line in lines:
                        d_ = json.loads(line)
                        relations = d_['relations']

                        # TODO - ensure only one id per type simplifier
                        set_ = dict((v['dst_id'], v) for v in relations).values()
                        relations = [_ for _ in set_]

                        if d_['name'] in ['ResearchStudy', 'research_study']:
                            # link the ResearchStudy to the gen3 project
                            relations.append({"dst_id": project_node_id, "dst_name": "Project", "label": "project"})
                            logger.info(
                                f"adding project relation from project({project_node_id}) to research_study{d_['id']}")

                        if len(relations) == 0:
                            msg = f"No relations for {d_['name']}"
                            if msg not in LOGGED_ALREADY:
                                LOGGED_ALREADY.append(msg)
                                print(msg)
                            continue

                        record_count += 1
                        for relation in relations:

                            # entity_name_underscore = inflection.underscore(entity_name)
                            dst_name_camel = inflection.camelize(relation['dst_name'])

                            edge_table_mapping = next(
                                iter(
                                    [
                                        m for m in mapping
                                        if m['srcclass'] == entity_name and m['dstclass'] == dst_name_camel
                                    ]
                                ),
                                None
                            )
                            if not edge_table_mapping and relation['dst_name'] in dependency_order:
                                msg = f"No mapping for src {entity_name} dst {relation['dst_name']}"
                                if msg not in LOGGED_ALREADY:
                                    logger.warning(msg)
                                    for m in mapping:
                                        logger.debug(m)
                                    LOGGED_ALREADY.append(msg)
                                continue
                            if not edge_table_mapping:
                                msg = f"No mapping for src {entity_name} dst {relation['dst_name']}"
                                if msg not in LOGGED_ALREADY:
                                    print(msg)
                                    LOGGED_ALREADY.append(msg)
                                continue
                            table_name = edge_table_mapping['tablename']
                            # print(f"Mapping for src {entity_name} dst {relation['dst_name']} {table_name} {edge_table_mapping}")
                            buf = buffers[table_name]
                            # src_id | dst_id | acl | _sysan | _props | created |
                            buf.write(f"{d_['id']}|{relation['dst_id']}|{{}}|{{}}|{{}}|{datetime.now()}\n")
                    for table_name, buf in buffers.items():
                        buf.seek(0)
                        # Creates temporary empty table with same columns and types as
                        # the final table
                        cursor.execute(
                            f"""
                            CREATE TEMPORARY TABLE "tmp_{table_name}" (LIKE "{table_name}")
                            ON COMMIT DROP
                            """
                        )

                        columns = ['src_id', 'dst_id', 'acl', '_sysan', '_props', 'created']
                        update_set = ", ".join([f"{v}=EXCLUDED.{v}" for v in ['acl', '_sysan', '_props', 'created']])
                        # efficient way to write to postgres
                        cursor.copy_from(buf, f"tmp_{table_name}", sep='|',
                                         columns=columns)
                        # handle conflicts
                        cursor.execute(
                            f"""
                            INSERT INTO "{table_name}" ({', '.join(columns)})
                            SELECT  {', '.join(columns)} FROM "tmp_{table_name}"
                            ON CONFLICT (src_id, dst_id) DO UPDATE SET {update_set}
                            """
                        )
                        logger.info(
                            f"wrote {record_count} records to {table_name} from {path} {entity_name} {relation['dst_name']}")
                        connection.commit()
        connection.commit()


def meta_upload(source_path, program, project, silent, dictionary_path, config_path,
                file_name_pattern='**/*.ndjson'):
    """Copy simplified json into Gen3."""
    assert pathlib.Path(source_path).is_dir(), f"{source_path} should be a directory"
    assert pathlib.Path(config_path).is_file(), f"{config_path} should be a file"
    assert dictionary_path, "dictionary_path cannot be empty"

    config_path = pathlib.Path(config_path)
    assert config_path.is_file()
    with open(config_path) as fp:
        gen3_config = yaml.load(fp, SafeLoader)

    dependency_order = [c for c in gen3_config['dependency_order'] if not c.startswith('_')]
    dependency_order = [c for c in dependency_order if c not in ['Program', 'Project']]

    ensure_project(program, project)

    # check db connection
    conn = _connect_to_postgres()
    assert conn
    logger.info("Connected to postgres")

    # check program/project exist
    cur = conn.cursor()
    cur.execute("select node_id, _props from \"node_program\";")
    programs = cur.fetchall()
    programs = [{'node_id': p[0], '_props': p[1]} for p in programs]
    _ = next(iter([p for p in programs if p['_props']['name'] == program]), None)
    assert _, f"{program} not found in node_program table"
    cur.execute("select node_id, _props from \"node_project\";")
    projects = cur.fetchall()
    projects = [{'node_id': p[0], '_props': p[1]} for p in projects]
    project_code = project
    project_node_id = next(iter([p['node_id'] for p in projects if p['_props']['code'] == project_code]), None)
    assert project_node_id, f"{project} not found in node_project"
    project_id = f"{program}-{project}"
    logger.info(f"Program and project exist: {project_id} {project_node_id}")

    # check files
    input_path = pathlib.Path(source_path)
    assert input_path.is_dir(), f"{input_path} should be a directory"
    files = [fn for fn in input_path.glob(file_name_pattern)]
    assert len(files) > 0, f"No files found at {input_path}/{file_name_pattern}"

    # check the mappings
    dictionary_dir = dictionary_path if 'http' not in dictionary_path else None
    dictionary_url = dictionary_path if 'http' in dictionary_path else None
    mappings = [mapping for mapping in _table_mappings(dictionary_dir, dictionary_url)]

    # load the files
    logger.info("Loading vertices")
    load_vertices(files, conn, dependency_order, project_id, mappings)

    logger.info("Loading edges")
    load_edges(files, conn, dependency_order, mappings, project_node_id)
    logger.info("Done")


# see https://github.com/uc-cdis/sheepdog/blob/master/sheepdog/globals.py#L51-L52
PROGRAM_SEED = uuid.UUID("85b08c6a-56a6-4474-9c30-b65abfd214a8")
PROJECT_SEED = uuid.UUID("249b4405-2c69-45d9-96bc-7410333d5d80")


def ensure_project(program, project) -> bool:
    """Ensure project exists in sheepdog database."""
    # check db connection
    conn = _connect_to_postgres()
    assert conn
    logger.info("Connected to postgres")

    # check program/project exist
    cur = conn.cursor()
    cur.execute("select node_id, _props from \"node_program\";")
    programs = cur.fetchall()
    programs = [{'node_id': p[0], '_props': p[1]} for p in programs]
    _ = next(iter([p for p in programs if p['_props']['name'] == program]), None)
    program_node_id = None
    if not _:  # program does not exist
        logger.info(f"Program {program} does not exist")
        program_node_id = str(uuid.uuid5(PROGRAM_SEED, program))
        cur.execute(
            "INSERT INTO node_program(node_id, _props) VALUES (%s, %s) ON CONFLICT DO NOTHING",
            (program_node_id, json.dumps({'name': program, 'type': 'program', "dbgap_accession_number": program}))
        )
        conn.commit()
        logger.info(f"Created Program {program}: {program_node_id}")
    else:
        program_node_id = _['node_id']
        logger.info(f"Program {program} exists: {program_node_id}")

    cur.execute("""
        select node_id, _props->>'code' as code  from node_project where node_id in (select src_id
        from
        edge_projectmemberofprogram
        where dst_id = (select node_id from node_program where _props->>'name' = %s)) and _props->>'code' = %s ;""",
                (program, project,)
                )
    project_node_id = None
    _ = cur.fetchone()
    if _:
        project_node_id, _ = _

    if not project_node_id:  # project does not exist
        logger.info(f"Project {project} does not exist")
        project_node_id = str(uuid.uuid5(PROJECT_SEED, project))
        cur.execute(
            "INSERT INTO node_project(node_id, _props) VALUES (%s, %s) ON CONFLICT DO NOTHING",
            (project_node_id,
             json.dumps({'code': project, 'type': 'project', "state": "open", "dbgap_accession_number": project}))
        )
        conn.commit()
        logger.info(f"Created Project {project}: {project_node_id}")
        cur.execute(
            "INSERT INTO edge_projectmemberofprogram(src_id, dst_id)  VALUES (%s, %s) ON CONFLICT DO NOTHING",
            (project_node_id, program_node_id)
        )
        conn.commit()
        logger.info(f"Created edge_projectmemberofprogram between {project_node_id} -> {program_node_id}")

    project_id = f"{program}-{project}"
    logger.info(f"Program and project exist: {project_id} {project_node_id}")


def empty_project(config_path, dictionary_path, program, project):
    """Remove all nodes from metadata graph."""

    config_path = pathlib.Path(config_path)
    assert config_path.is_file()
    with open(config_path) as fp:
        gen3_config = yaml.load(fp, SafeLoader)

    project_id = f"{program}-{project}"
    logger.info(f"Emptying project {project_id}")
    dependency_order = [c for c in gen3_config['dependency_order'] if not c.startswith('_')]
    dependency_order = [c for c in dependency_order if c not in ['Program', 'Project']]

    dictionary_dir = dictionary_path if 'http' not in dictionary_path else None
    dictionary_url = dictionary_path if 'http' in dictionary_path else None
    mappings = [mapping for mapping in _table_mappings(dictionary_dir, dictionary_url)]

    conn = _connect_to_postgres()
    for entity_name in dependency_order:
        data_table_name = next(
            iter(
                set([m['dsttable'] for m in mappings if m['dstclass'].lower() == entity_name.lower()] +
                    [m['srctable'] for m in mappings if m['srcclass'].lower() == entity_name.lower()])
            ),
            None)
        if not data_table_name:
            logger.warning(f"No mapping found for {entity_name} skipping")
            continue
        logger.info(f"Truncating {data_table_name} for {project_id}")
        with conn.cursor() as cursor:
            cursor.execute(f"DELETE FROM {data_table_name} WHERE _props->>'project_id' = %s", (project_id,))
            conn.commit()

    conn.commit()
    logger.info(f"Done emptying project {project_id}")



import logging
import pathlib
from typing import Iterator

import fastjsonschema
import inflection
import orjson
import requests
from dictionaryutils import DataDictionary
from jsonschema.exceptions import ValidationError

from aced_submission.util import _to_file, ParseResult

logger = logging.getLogger(__name__)

LOGGED_ALREADY = []
COMPILED_SCHEMAS = {}


def log_once(msg):
    if msg not in LOGGED_ALREADY:
        logger.info(msg)
        LOGGED_ALREADY.append(msg)


def _validate(gen3_resource: dict, schemas: dict):
    """Ensure resource is valid for gen3 schema."""
    try:
        resource = None
        for expected_property in ['id', 'relations', 'object']:
            assert expected_property in gen3_resource, f"{expected_property} not in {gen3_resource}"

        resource = gen3_resource['object']
        resource_type = resource['resourceType']
        key = inflection.underscore(resource_type)
        schema = schemas.get(key, schemas.get(f"{key}.yaml", None))
        assert schema, f"Could not find schema for {key}"
        actual_keys = set(resource.keys())
        expected_keys = set(schema['properties'].keys()).union(set([_['name'] for _ in schema['links']]))
        if not actual_keys.issubset(expected_keys):
            if not schema.get('additionalProperties', False):
                assert False, f"Is not a subset {actual_keys - expected_keys} not expected"
            else:
                for _ in actual_keys - expected_keys:
                    msg = f"extra property: {resource_type}.{_} {type(resource[_]).__name__}"
                    log_once(msg)
        if 'required' in schema:
            for _ in schema['required']:
                assert _ in resource, f"{_} missing {resource}"
        compiled_schema = COMPILED_SCHEMAS.get(key, None)
        if not compiled_schema:
            compiled_schema = fastjsonschema.compile(
                schema,
                formats={
                    'time': r'^(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$',
                    'date': r'^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])?$',
                    'binary': r'^(?:[A-Za-z0-9+/]{4})*(?:[A-Za-z0-9+/]{2}==|[A-Za-z0-9+/]{3}=)?$',
                    "date-time": r'^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$',
                    "uri": r'\w+:(\/?\/?)[^\s]+',
                    "uuid": r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-5][0-9a-f]{3}-[089ab][0-9a-f]{3}-[0-9a-f]{12}$"
                }
            )
            COMPILED_SCHEMAS[key] = compiled_schema
        compiled_schema(resource)
        return ParseResult(resource=resource, exception=None, path=None, resource_id=resource.get('id', None))
    except (ValidationError, AssertionError) as e:
        if resource:
            return ParseResult(resource=None, exception=e, path=None, resource_id=resource.get('id', None))
        else:
            return ParseResult(resource=None, exception=e, path=None)


def directory_reader(
        directory_path: pathlib.Path,
        schema_path: str,
        pattern: str = '*.ndjson',
        validate=True) -> Iterator[ParseResult]:
    """Extract FHIR resources from directory"""

    assert directory_path.is_dir(), f"{directory_path.name} is not a directory"

    if 'http' in schema_path:
        schemas = requests.get(schema_path).json()
    else:
        schemas = DataDictionary(local_file=schema_path).schema

    input_files = [_ for _ in directory_path.glob(pattern)]
    for input_file in input_files:
        logger.info(input_file)
        if not input_file.is_file():
            continue
        fp = _to_file(input_file)
        with fp:
            offset = 0
            for line in fp.readlines():
                gen3_resource = orjson.loads(line)
                parse_result = _validate(gen3_resource, schemas=schemas)
                parse_result.path = input_file
                parse_result.offset = offset
                offset += 1
                yield parse_result



import logging
import pathlib
from typing import TextIO

import click


class NaturalOrderGroup(click.Group):
    """See https://github.com/pallets/click/issues/513."""
    def list_commands(self, ctx):
        return self.commands.keys()


class EmitterContextManager:
    """Maintain file pointers to output directory."""

    def __init__(self, output_path: str, verbose=False, file_mode="w",
                 logger=logging.getLogger("EmitterContextManager")):
        """Ensure output_path exists, init emitter dict."""
        output_path = pathlib.Path(output_path)
        if not output_path.exists():
            output_path.mkdir(parents=True)
        assert output_path.is_dir(), f"{output_path} not a directory?"

        self.output_path = output_path
        """destination directory"""
        self.emitters = {}
        """open file pointers"""
        self.verbose = verbose
        """log activity"""
        self.file_mode = file_mode
        """mode for file opens"""
        self.logger = logger

    def __enter__(self):
        """Ensure output_path exists, init emitter dict.
        """
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Close all open files."""
        for _ in self.emitters.values():
            _.close()
            if self.verbose:
                self.logger.info(f"wrote {_.name}")

    def emit(self, name: str) -> TextIO:
        """Maintain a hash of open files."""
        if name not in self.emitters:
            self.emitters[name] = open(self.output_path / f"{name}.ndjson", self.file_mode)
            if self.verbose:
                self.logger.debug(f"opened {self.emitters[name].name}")
        return self.emitters[name]



"""Adds, retrieves, and deletes FHIR resources in a FHIR store."""
import json
import pathlib
import sys

import click
import opensearchpy as elasticsearch
import yaml
from opensearchpy import OpenSearch as Elasticsearch

from aced_submission import NaturalOrderGroup
from aced_submission.meta_flat_load import read_ndjson, write_bulk_http, DEFAULT_ELASTIC


@click.group(cls=NaturalOrderGroup, name='fhir')
def fhir_store():
    """Adds, retrieves, and deletes FHIR resources in a FHIR store."""
    pass


def resource_generator(project_id, file_path):
    """Yield FHIR resources from a ndjson file."""
    program, project = project_id.split('-')
    assert program, "program is required"
    assert project, "project is required"

    for _ in read_ndjson(file_path):
        assert 'id' in _, f"resource {_} does not have an 'id'"
        assert 'resourceType' in _, f"resource {_} does not have a 'resourceType'"
        _["auth_resource_path"] = f"/programs/{program}/projects/{project}"
        yield _


def fhir_put(project_id, path, elastic_url) -> list[str]:
    """Upsert FHIR resources to a FHIR store."""
    assert project_id.count('-') == 1, f"{project_id} should have a single '-' separating program and project"

    elastic = Elasticsearch([elastic_url], request_timeout=120)

    index = doc_type = 'fhir'
    limit = None
    logs = []
    for file_path in pathlib.Path(path).glob('*.ndjson'):

        write_bulk_http(elastic=elastic, index=index, doc_type=doc_type, limit=limit,
                        generator=resource_generator(project_id, file_path))

        logs.append(f"wrote {file_path} to elasticsearch/{index}")

    return logs


def fhir_get(project_id, path, elastic_url) -> list[str]:
    """Retrieve FHIR resources from FHIR store, write to path/resourceType.ndjson."""
    assert project_id.count('-') == 1, f"{project_id} should have a single '-' separating program and project"
    program, project = project_id.split('-')
    assert program, "program is required"
    assert project, "project is required"

    elastic = Elasticsearch([elastic_url], request_timeout=120)

    index = 'fhir'
    logs = []

    emitters = {}
    open_files = []

    pathlib.Path(path).mkdir(parents=True, exist_ok=True)

    def _emitter(_resource_type):
        """Maintain has of open files."""
        if resource_type not in emitters:
            file_path = pathlib.Path(path) / f"{_resource_type}.ndjson"
            emitters[_resource_type] = file_path.open('w')
            open_files.append(file_path)
        return emitters[_resource_type]

    auth_resource_path = f"/programs/{program}/projects/{project}"

    for _ in elasticsearch.helpers.scan(
        client=elastic,
        query={
          "query": {
            "term": {
              "auth_resource_path.keyword": {
                "value": auth_resource_path
              }
            }
          }
        },
        index=index
    ):
        resource_type = _['_source']['resourceType']
        _file = _emitter(resource_type)
        del _['_source']['auth_resource_path']
        json.dump(_['_source'], _file)
        _file.write('\n')

    for file in emitters.values():
        file.close()
    for file in open_files:
        logs.append(f"wrote {file}")

    return logs


def fhir_delete(project_id, elastic_url) -> list[str]:
    """Delete FHIR resources from FHIR store based on project_id."""
    assert project_id.count('-') == 1, f"{project_id} should have a single '-' separating program and project"
    program, project = project_id.split('-')
    assert program, "program is required"
    assert project, "project is required"

    elastic = Elasticsearch([elastic_url], request_timeout=120)

    index = 'fhir'
    logs = []

    auth_resource_path = f"/programs/{program}/projects/{project}"

    deleted_count = 0
    actions = []

    for _ in elasticsearch.helpers.scan(
        client=elastic,
        query={
            "query": {
                "term": {
                    "auth_resource_path.keyword": {
                        "value": auth_resource_path
                    }
                }
            }
        },
        index=index
    ):
        actions.append({
            "_op_type": "delete",
            "_index": index,
            "_id": _['_id']
        })

    if actions:
        try:
            elasticsearch.helpers.bulk(elastic, actions)
            deleted_count += len(actions)
        except elasticsearch.ElasticsearchException as e:
            logs.append(f"Error deleting resources: {e}")

    logs.append(f"Deleted {deleted_count} resources for project {project_id}")

    return logs


@fhir_store.command(name='put')
@click.option('--project_id', required=True, show_default=True,
              help="Gen3 program-project")
@click.option('--format', 'output_format',
              default='yaml',
              show_default=True,
              type=click.Choice(['yaml', 'json'], case_sensitive=False))
@click.option('--elastic_url', default=DEFAULT_ELASTIC, show_default=True)
@click.argument('path', default=None, required=True)
def _fhir_put(project_id, output_format, path, elastic_url):
    """Upsert FHIR resources to a FHIR store.

    \b
    PATH: directory containing ndjson files
    """
    logs = fhir_put(project_id, path, elastic_url)
    if output_format == 'yaml':
        yaml.dump(logs, sys.stdout, default_flow_style=False)
    else:
        json.dump(logs, sys.stdout, indent=2)


@fhir_store.command(name='get')
@click.option('--project_id', required=True, show_default=True,
              help="Gen3 program-project")
@click.option('--format', 'output_format',
              default='yaml',
              show_default=True,
              type=click.Choice(['yaml', 'json'], case_sensitive=False))
@click.option('--elastic_url', default=DEFAULT_ELASTIC, show_default=True)
@click.argument('path', default=None, required=True)
def _fhir_get(project_id, output_format, path, elastic_url):
    """Exports all resources for project_id to a directory.
    \b
    PATH: directory to write ndjson files
    """
    logs = fhir_get(project_id, path, elastic_url)
    if output_format == 'yaml':
        yaml.dump(logs, sys.stdout, default_flow_style=False)
    else:
        json.dump(logs, sys.stdout, indent=2)


@fhir_store.command(name='delete')
@click.option('--project_id', required=True, show_default=True,
              help="Gen3 program-project")
@click.option('--format', 'output_format',
              default='yaml',
              show_default=True,
              type=click.Choice(['yaml', 'json'], case_sensitive=False))
@click.option('--elastic_url', default=DEFAULT_ELASTIC, show_default=True)
def _fhir_delete(project_id, output_format, elastic_url):
    """Deletes all resources for project_id"""
    logs = fhir_delete(project_id, elastic_url)
    if output_format == 'yaml':
        yaml.dump(logs, sys.stdout, default_flow_style=False)
    else:
        json.dump(logs, sys.stdout, indent=2)



"""Load flat indexes into elasticsearch."""


import csv
import json
import logging
import os
import pathlib
import sqlite3
import tempfile
import uuid
from datetime import datetime
from functools import lru_cache
from itertools import islice
from typing import Dict, Iterator, Any, Generator, List

import click
from dateutil.parser import parse
from dateutil import tz
from gen3_tracker.meta.dataframer import LocalFHIRDatabase
import orjson
from opensearchpy import OpenSearch as Elasticsearch
from opensearchpy.helpers import streaming_bulk

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
logging.getLogger('elasticsearch').setLevel(logging.WARNING)

DEFAULT_ELASTIC = "http://localhost:9200"

k8s_elastic = os.environ.get('GEN3_ELASTICSEARCH_MASTER_PORT', None)
if k8s_elastic:
    DEFAULT_ELASTIC = f"http://{k8s_elastic.replace('tcp://', '')}"

# TODO - fix me should be gen3.aced-idp.org but we need to coordinate with gitops.json
ES_INDEX_PREFIX = "gen3.aced.io"

ACED_NAMESPACE = uuid.uuid3(uuid.NAMESPACE_DNS, 'aced-ipd.org')


def create_id(key: str) -> str:
    """Create an idempotent ID from the input string."""
    return str(uuid.uuid5(ACED_NAMESPACE, key))


def read_ndjson(path: str) -> Iterator[Dict]:
    """Read ndjson file, load json line by line."""
    with open(path) as jsonfile:
        for l_ in jsonfile.readlines():
            yield json.loads(l_)


def read_tsv(path: str) -> Iterator[Dict]:
    """Read tsv file line by line."""
    with open(path) as tsv_file:
        reader = csv.DictReader(tsv_file, delimiter="\t")
        for row in reader:
            yield row


def generate_elasticsearch_mapping(df: List[Dict]) -> Dict[str, Any]:
    """
    Generates an Elasticsearch mapping from a "DataFrame".

    Args:
        df (Dict): A list of dict.

    Returns:
        Dict[str, Any]: The generated Elasticsearch mapping.
    """

    def is_integer_dtype(value: Any) -> bool:
        return isinstance(value, int) and not isinstance(value, bool)

    def is_float_dtype(value: Any) -> bool:
        return isinstance(value, float)

    def is_bool_dtype(value: Any) -> bool:
        return isinstance(value, bool)

    def is_datetime64_any_dtype(value: Any) -> bool:
        try:
            if not isinstance(value, str):
                raise ValueError('Value is not a string')
            tzinfos = {"PDT": tz.gettz('US/West')}
            parse(value, fuzzy=True, tzinfos=tzinfos)
            return True
        except Exception:  # noqa
            return False

    def is_object_dtype(value: Any) -> bool:
        return isinstance(value, list) or isinstance(value, dict)

    dynamic_templates = [
         {
             "strings": {
                 "match_mapping_type": "string",
                 "mapping": {
                     "type": "keyword"
                 }
             }
         }
    ]
    mapping = {"mappings": {"properties": {}, 'dynamic_templates': dynamic_templates}}
    # see https://github.com/uc-cdis/guppy/blob/f5bb705dae6f3417e471ba2e43ce4b61ba5026fb/src/server/schema.js#L4
    # for the guppy schema types
    for row in df:
        for column in row.keys():
            if is_integer_dtype(row[column]):
                mapping["mappings"]["properties"][column] = {"type": "integer"}
            elif is_float_dtype(row[column]):
                mapping["mappings"]["properties"][column] = {"type": "float"}
            elif is_bool_dtype(row[column]):
                mapping["mappings"]["properties"][column] = {"type": "keyword"}
            elif is_datetime64_any_dtype(row[column]):
                mapping["mappings"]["properties"][column] = {"type": "text"}
            elif is_object_dtype(row[column]):
                if isinstance(row[column], list):
                    mapping["mappings"]["properties"][column] = {"type": "keyword"}
                else:
                    mapping["mappings"]["properties"][column] = {"type": "keyword"}
            elif isinstance(row[column], str):
                mapping["mappings"]["properties"][column] = {"type": "keyword"}
    return mapping


def write_array_aliases(doc_type, alias, elastic=DEFAULT_ELASTIC, name_space=ES_INDEX_PREFIX):
    """Write the array aliases."""
    # EXPECTED_ALIASES = {
    #     ".kibana_1": {
    #         "aliases": {
    #             ".kibana": {}
    #         }
    #     },
    #     "etl-array-config_0": {
    #         "aliases": {
    #             "etl-array-config": {},
    #             "etl_array-config": {},
    #             "time_2022-08-25T01:44:47.115494": {}
    #         }
    #     },
    #     "etl_0": {
    #         "aliases": {
    #             "etl": {},
    #             "time_2022-08-25T01:44:47.115494": {}
    #         }
    #     },
    #     "file-array-config_0": {
    #         "aliases": {
    #             "file-array-config": {},
    #             "file_array-config": {},
    #             "time_2022-08-25T01:44:47.115494": {}
    #         }
    #     },
    #     "file_0": {
    #         "aliases": {
    #             "file": {},
    #             "time_2022-08-25T01:44:47.115494": {}
    #         }
    #     }
    # }
    return {
        "method": 'POST',
        "url": f'{elastic}/_aliases',
        "json": {
            "actions": [
                {"add": {"index": f"{name_space}_{doc_type}-array-config_0",
                         "alias": f"{name_space}_array-config"}},
                {"add": {"index": f"{name_space}_{doc_type}-array-config_0",
                         "alias": f"{alias}_array-config"}}
            ]}
    }


def write_array_config(doc_type, alias, field_array, elastic=DEFAULT_ELASTIC, name_space=ES_INDEX_PREFIX):
    """Write the array config."""
    return {
        "method": 'PUT',
        "url": f'/{name_space}_{doc_type}-array-config_0/_doc/{alias}',
        "json": {"timestamp": datetime.now().isoformat(), "array": field_array}
    }


def write_alias_config(doc_type, alias, elastic=DEFAULT_ELASTIC, name_space=ES_INDEX_PREFIX):
    """Write the alias config."""
    return {
        "method": 'POST',
        "url": f'{elastic}/_aliases',
        "json": {"actions": [{"add": {"index": f"{name_space}_{doc_type}_0", "alias": alias}}]}
    }


def write_bulk_http(elastic, index, limit, doc_type, generator) -> None:
    """Use efficient method to write to elastic, assumes a)generator is a list of dictionaries b) indices already exist. """
    counter = 0

    def _bulker(generator_, counter_=counter):
        for dict_ in generator_:
            if limit and counter_ > limit:
                break  # for testing
            yield {
                '_index': index,
                '_op_type': 'index',
                '_source': dict_,
                # use the id from the FHIR object to upsert information
                '_id': dict_['id']
            }
            counter_ += 1
            if counter_ % 10000 == 0:
                logger.info(f"{counter_} records written")
        logger.info(f"{counter_} records written")

    logger.info(f'Writing bulk to {index} limit {limit}.')
    _ = streaming_bulk(client=elastic,
                       actions=(d for d in _bulker(generator)),
                       request_timeout=120,
                       max_retries=5)

    return


def observation_generator(project_id, generator) -> Iterator[Dict]:
    """Render guppy index for observation."""
    program, project = project_id.split('-')
    for observation in generator:
        observation['project_id'] = project_id
        observation["auth_resource_path"] = f"/programs/{program}/projects/{project}"
        yield observation


def file_generator(project_id, generator) -> Iterator[Dict]:
    """Render guppy index for file."""
    program, project = project_id.split('-')
    for file in generator:
        file['project_id'] = project_id
        file["auth_resource_path"] = f"/programs/{program}/projects/{project}"
        yield file


@lru_cache(maxsize=1024 * 10)
def fetch_denormalized_patient(connection, patient_id):
    """Retrieve unique conditions and family history"""

    fh_condition = []
    fh_condition_coding = []
    condition = []
    condition_coding = []
    patient = None

    for row in connection.execute('select entity from patient where id = ? limit 1', (patient_id,)):
        patient = orjson.loads(row[0])
        break

    for row in connection.execute('select entity from family_history where patient_id = ? ', (patient_id,)):
        family_history = orjson.loads(row[0])
        for _ in family_history['condition']:
            if _ not in fh_condition:
                fh_condition.append(_)
        for _ in family_history['condition_coding']:
            if _ not in fh_condition_coding:
                fh_condition_coding.append(_)

    for row in connection.execute('select entity from condition where patient_id = ? ', (patient_id,)):
        condition_ = orjson.loads(row[0])
        if condition_['code'] not in condition:
            condition.append(condition_['code'])
            condition_coding.append(condition_['code_coding'])

    return {
        'patient': patient, 'condition': condition, 'condition_coding': condition_coding,
        'fh_condition': fh_condition, 'fh_condition_coding': fh_condition_coding
    }


def setup_aliases(alias, doc_type, elastic, field_array, index):
    """Create the alias to the data index"""
    if not elastic.indices.exists_alias(alias):
        logger.warning(f"Creating alias {alias}.")
        elastic.indices.put_alias(index, alias)
    else:
        logger.info(f"Alias {alias} already exists.")
    # create a configuration index that guppy will read that describes the array fields
    # TODO - find a doc or code reference in guppy that explains how this is used
    array_config_index = f'{ES_INDEX_PREFIX}_{doc_type}-array-config_0'
    try:
        mapping = {
            "mappings": {
                "properties": {
                    "timestamp": {"type": "date"},
                    "array": {"type": "keyword"},
                }
            }
        }
        if not elastic.indices.exists(index=array_config_index):
            logger.warning(f"Creating index {array_config_index}.")
            elastic.indices.create(index=array_config_index, body=mapping)

            elastic.indices.update_aliases(
                {"actions": [{"add": {"index": f"{ES_INDEX_PREFIX}_{doc_type}_0", "alias": alias}}]}
            )
            elastic.indices.update_aliases({
                "actions": [
                    {"add": {"index": f"{ES_INDEX_PREFIX}_{doc_type}-array-config_0",
                             "alias": f"{ES_INDEX_PREFIX}_array-config"}},
                    {"add": {"index": f"{ES_INDEX_PREFIX}_{doc_type}-array-config_0",
                             "alias": f"{doc_type}_array-config"}}
                ]}
            )
            logger.warning(f"Updated aliases {array_config_index}")
        else:
            logger.warning(f"{array_config_index} already exists.")
        elastic.index(index=array_config_index, id=alias,
                      body={"timestamp": datetime.now().isoformat(), "array": field_array},
                      refresh='wait_for')
        logger.warning(f"Populated {array_config_index} field_array {field_array}")

    except Exception as e:
        logger.warning(f"Could not create index. {array_config_index} {str(e)}")
        logger.warning("Continuing to load.")


@click.group('flat')
def cli():
    """Load flat indexes into elasticsearch."""
    pass


def write_flat_file(output_path, index, doc_type, limit, generator):
    """Write the flat model to a file."""
    counter_ = 0
    pathlib.Path(output_path).mkdir(parents=True, exist_ok=True)
    with open(f"{output_path}/{doc_type}.ndjson", "wb") as fp:
        for dict_ in generator:
            fp.write(
                orjson.dumps(
                    {
                        'id': dict_['id'],
                        'object': dict_,
                        'name': doc_type,
                        'relations': []
                    }
                )
            )
            fp.write(b'\n')

            counter_ += 1
            if counter_ % 10000 == 0:
                logger.info(f"{counter_} records written")
        logger.info(f"{counter_} records written")


@cli.command('denormalize-patient')
@click.option('--input_path', required=True,
              default=None,
              show_default=True,
              help='Path to flattened json'
              )
def _denormalize_patient(input_path):
    denormalize_patient(input_path)


def denormalize_patient(input_path):
    """Gather Patient, FamilyHistory, Condition into sqlite db."""

    path = pathlib.Path(input_path)

    def _load_vertex(file_name):
        """Get the object and patient id"""
        if not (path / file_name).is_file():
            return
        for _ in read_ndjson(path / file_name):
            patient_id = None
            if len(_['relations']) == 1 and _['relations'][0]['dst_name'] == 'Patient':
                patient_id = _['relations'][0]['dst_id']
            _ = _['object']
            _['id'] = _['id']
            if patient_id:
                _['patient_id'] = patient_id
            yield _

    connection = sqlite3.connect('denormalized_patient.sqlite')
    with connection:
        connection.execute('DROP table IF EXISTS patient')
        connection.execute('DROP table IF EXISTS family_history')
        connection.execute('DROP table IF EXISTS condition')
        connection.execute('CREATE TABLE if not exists patient (id PRIMARY KEY, entity Text)')
        connection.execute('CREATE TABLE if not exists family_history (id PRIMARY KEY, patient_id Text, entity Text)')
        connection.execute('CREATE TABLE if not exists condition (id PRIMARY KEY, patient_id Text, entity Text)')
    with connection:
        connection.executemany('insert into patient values (?, ?)',
                               [(entity['id'], orjson.dumps(entity).decode(),) for entity in
                                _load_vertex('Patient.ndjson')])
    with connection:
        connection.executemany('insert into family_history values (?, ?, ?)',
                               [(entity['id'], entity['patient_id'], orjson.dumps(entity).decode(),) for entity in
                                _load_vertex('FamilyMemberHistory.ndjson')])
    with connection:
        connection.executemany('insert into condition values (?, ?, ?)',
                               [(entity['id'], entity['patient_id'], orjson.dumps(entity).decode(),) for entity in
                                _load_vertex('Condition.ndjson')])
    with connection:
        connection.execute('CREATE INDEX if not exists condition_patient_id on condition(patient_id)')
        connection.execute('CREATE INDEX if not exists family_history_patient_id on condition(patient_id)')


def compare_mapping(existing_mapping: Dict[str, Any], new_mapping: Dict[str, Any]) -> Dict[str, Any]:
    """
    Compares an existing Elasticsearch index mapping with a new mapping create update for the index by adding missing fields.

    Args:
        existing_mapping (Dict[str, Any]): The existing mapping.
        new_mapping (Dict[str, Any]): The new mapping to compare against the existing mapping.

    Returns:
        None
    """

    new_properties = new_mapping['mappings']['properties']
    existing_properties = existing_mapping['mappings']['properties']

    # Find differences and update mapping
    updates = {}
    for field, field_type in new_properties.items():
        if field not in existing_properties:
            updates[field] = field_type

    return updates


def ndjson_file_generator(path):
    """Read ndjson file line by line."""
    with open(path) as f:
        for l_ in f.readlines():
            yield orjson.loads(l_)


@cli.command('load')
@click.option('--input_path', required=True,
              default='META/',
              show_default=True,
              help='Path to flattened json'
              )
@click.option('--project_id', required=True,
              default=None,
              show_default=True,
              help='program-project'
              )
@click.option('--data_type', required=True,
              default='observation',
              type=click.Choice(['observation', 'file'], case_sensitive=False),
              show_default=True,
              help='index to load[observation, file] '
              )
def _load_flat(input_path, project_id, data_type):
    import tempfile

    work_path = tempfile.TemporaryDirectory(delete=False).name
    assert pathlib.Path(work_path).exists(), f"Directory {work_path} does not exist."
    work_path = pathlib.Path(work_path)
    db_path = (work_path / "local_fhir.db")
    db_path.unlink(missing_ok=True)

    db = LocalFHIRDatabase(db_name=db_path)
    db.load_ndjson_from_dir(path=input_path)

    load_flat(project_id=project_id,
              generator=db.flattened_observations(),
              index=data_type,
              limit=None,
              elastic_url=DEFAULT_ELASTIC,
              output_path=None
              )


def load_flat(project_id: str, index: str, generator: Generator[dict, None, None], limit: str, elastic_url: str, output_path: str):
    """Loads flattened FHIR data into Elasticsearch database. Replaces tube-lite"""

    assert index in ["observation", "file"], f"Index {index} does not have a supported generator"

    if limit:
        limit = int(limit)

    elastic = Elasticsearch([elastic_url], request_timeout=120)
    assert elastic.ping(), f"Connection to {elastic_url} failed"
    index = index.lower()

    def load_index(elastic: Elasticsearch, output_path: str, es_index: str, alias: str, doc_type: str, limit: int):
        if not output_path:
            # create the index and write data into it.

            # since we need to read the generator twice, once to create the indices and once to write the data to ES
            # Get the path of the temporary file
            temp_path = tempfile.NamedTemporaryFile(delete=False).name
            # just write the data to it
            with open(temp_path, mode='w') as f:
                for _ in generator:
                    f.write(orjson.dumps(_).decode())
                    f.write('\n')

            if index == "observation":
                loading_generator = observation_generator(project_id, ndjson_file_generator(temp_path))
            elif index == "file":
                loading_generator = file_generator(project_id, ndjson_file_generator(temp_path))

            if elastic.indices.exists(index=es_index):
                logger.info(f"Index {es_index} exists.")

                existing_mapping = elastic.indices.get_mapping(index=es_index)
                assert es_index in existing_mapping, f"doc_type {es_index} not in {existing_mapping}"
                existing_mapping = existing_mapping[es_index]

                new_mapping = generate_elasticsearch_mapping(ndjson_file_generator(temp_path))
                updates = compare_mapping(existing_mapping, new_mapping)
                if updates != {}:
                    update_body = {
                        "properties": updates
                    }
                    elastic.indices.put_mapping(index=es_index, body=update_body)
                    logger.info(f"Updated {es_index} with {updates}")
                else:
                    logger.info(f"No updates needed for {es_index}")
            else:
                logger.info(f"Index {es_index} does not exist.")
                mapping = generate_elasticsearch_mapping(ndjson_file_generator(temp_path))
                elastic.indices.create(index=es_index, body=mapping)
                logger.info(f"Created {es_index}")

            write_bulk_http(elastic=elastic, index=es_index, doc_type=doc_type, limit=limit,
                            generator=loading_generator)

            field_array = set()
            for _ in ndjson_file_generator(temp_path):
                field_array.update([k for k, v in _.items() if isinstance(v, list)])
            field_array = list(field_array)
            setup_aliases(alias, doc_type, elastic, field_array, es_index)

            pathlib.Path(temp_path).unlink()

        else:
            # write file path
            write_flat_file(output_path=output_path, index=es_index, doc_type=doc_type, limit=limit,
                            generator=loading_generator)

    if index == 'observation':
        doc_type = 'observation'
        load_index(elastic=elastic, output_path=output_path, es_index=f"{ES_INDEX_PREFIX}_{doc_type}_0", alias="observation", doc_type=doc_type, limit=limit)

    elif index == 'file':
        doc_type = 'file'
        load_index(elastic=elastic, output_path=output_path, es_index=f"{ES_INDEX_PREFIX}_{doc_type}_0", alias="file", doc_type=doc_type, limit=limit)


def chunk(arr_range, arr_size):
    """Iterate in chunks."""
    arr_range = iter(arr_range)
    return iter(lambda: tuple(islice(arr_range, arr_size)), ())


@cli.command('counts')
@click.option('--project_id', required=True,
              default=None,
              show_default=True,
              help='program-project'
              )
def _counts(project_id):
    counts(project_id)


def counts(project_id):
    """Count the number of patients, observations, and files."""
    elastic = Elasticsearch([DEFAULT_ELASTIC], request_timeout=120)
    program, project = project_id.split('-')
    assert program, "program is required"
    assert project, "project is required"
    query = {
        "query": {
            "match": {
                "auth_resource_path": f"/programs/{program}/projects/{project}"
            }
        }
    }
    for index in ['observation', 'file']:
        # index = f"{ES_INDEX_PREFIX}_{index}_0"
        print(index, elastic.count(index=index, body=query)['count'])


@cli.command('rm')
@click.option('--project_id', required=True,
              default=None,
              show_default=True,
              help='program-project'
              )
@click.option('--index', required=True,
              default=None,
              show_default=True,
              help='one of observation, file'
              )
def _delete(project_id, index):
    delete(project_id, index)


def delete(project_id, index):
    """Delete items from elastic index for project_id."""
    elastic = Elasticsearch([DEFAULT_ELASTIC], request_timeout=120)
    assert project_id, "project_id is required"
    program, project = project_id.split('-')
    assert program, "program is required"
    assert project, "project is required"
    assert index, "index is required"
    query = {
        "query": {
            "match": {
                "auth_resource_path": f"/programs/{program}/projects/{project}"
            }
        }
    }
    print("deleting, waiting up to 5 min. for response")
    print(index, elastic.delete_by_query(index=index, body=query, timeout='5m'))


if __name__ == '__main__':
    cli()



import os
import pathlib
import jwt
import logging
import requests
from gen3.auth import Gen3Auth
from gen3.metadata import Gen3Metadata


# copied from gen3_util/config/__init__.py
def ensure_auth(refresh_file: [pathlib.Path, str] = None, validate: bool = False) -> Gen3Auth:
    """Confirm connection to Gen3 using their conventions.

    Args:
        refresh_file (pathlib.Path): The file containing the downloaded JSON web token.
        validate: check the connection by getting a new token

    """

    try:
        if refresh_file:
            if isinstance(refresh_file, str):
                refresh_file = pathlib.Path(refresh_file)
            auth = Gen3Auth(refresh_file=refresh_file.name)
        elif 'ACCESS_TOKEN' in os.environ:
            auth = Gen3Auth(refresh_file=f"accesstoken:///{os.getenv('ACCESS_TOKEN')}")
        else:
            auth = Gen3Auth()

        if validate:
            api_key = auth.refresh_access_token()
            assert api_key, "refresh_access_token failed"

    except (requests.exceptions.ConnectionError, AssertionError) as e:
        msg = (f"Could not get access."
               "See https://bit.ly/3NbKGi4, or, "
               "store the file in ~/.gen3/credentials.json or specify location with env GEN3_API_KEY "
               f"{e}")

        logging.getLogger(__name__).error(msg)
        raise AssertionError(msg)

    return auth


def discovery_get(project_id: str):
    """Fetches project information from discovery metadata-service"""

    auth = ensure_auth()
    discovery_client = Gen3Metadata(auth.endpoint, auth)

    try:
        data = discovery_client.get(project_id)
    except requests.exceptions.HTTPError as e:
        print(str(e))
        if e.response.status_code == 404:
            return {}
        return None

    return data


def discovery_delete(project_id: str):
    """Deletes project information to discovery metadata-service"""

    auth = ensure_auth()
    discovery_client = Gen3Metadata(auth.endpoint, auth)

    try:
        discovery_client.delete(project_id)
        print(f"Deleted {project_id}")
    except requests.exceptions.HTTPError as e:
        print(str(e))


def discovery_load(project_id: str, _subjects_count: int, description: str, location: str):
    """Writes project information to discovery metadata-service.
       Overwrites existing data"""

    program, project = project_id.split("-")
    auth = ensure_auth()
    token = auth.get_access_token()

    # Decode the jwt ACCESS_TOKEN to get the commons endpoint
    decoded_token = jwt.decode(token, secret=None, algorithms=["RS256"], options={"verify_signature": False})
    commons_url = decoded_token["iss"].removesuffix("/user").removeprefix("https://")

    discovery_client = Gen3Metadata(auth.endpoint, auth)
    gen3_discovery = {'tags': [
        {"name": program, "category": "Program"},
        {"name": project, "category": "Project"},
        {"name": project_id, "category": "Study Registration"},
        {"name": location, "category": "Study Location"},

    ], 'name': project, 'full_name': project, 'study_description': description}

    gen3_discovery['commons'] = "ACED"
    gen3_discovery['commons_name'] = "ACED Commons"
    gen3_discovery['commons_url'] = commons_url
    gen3_discovery['__manifest'] = 0
    gen3_discovery['_research_subject_count'] = int(_subjects_count)
    gen3_discovery['_unique_id'] = project_id
    gen3_discovery['study_id'] = project_id
    discoverable_data = dict(_guid_type="discovery_metadata", gen3_discovery=gen3_discovery)

    try:
        discovery_client.create(project_id, discoverable_data, aliases=None, overwrite=True)
        print(f"Added {project_id}")
    except requests.exceptions.HTTPError as e:
        print(str(e))
